from __future__ import unicode_literals
from base64 import standard_b64encode
import copy
import binascii
from libnacl.secret import SecretBox
from pymacaroons import Caveat

from pymacaroons.binders import HashSignaturesBinder
from pymacaroons.serializers.binary_serializer import BinarySerializer
from pymacaroons.exceptions import MacaroonInitException
from pymacaroons.utils import convert_to_bytes, convert_to_string, \
    create_initial_macaroon_signature, truncate_or_pad, generate_derived_key, \
    sign_third_party_caveat, sign_first_party_caveat


class Macaroon(object):

    def __init__(self,
                 location=None,
                 identifier=None,
                 key=None,
                 default_binder_class=None,
                 default_serializer=None,
                 caveats=None,
                 signature=None):
        self.caveats = caveats or []
        if location:
            self.location = location
        if identifier:
            self.identifier = identifier
        if location and identifier and key:
            self.signature = create_initial_macaroon_signature(
                convert_to_bytes(key), convert_to_bytes(identifier)
            )
        if signature:
            self.signature = signature
        self.binder_class = default_binder_class or HashSignaturesBinder
        self.serializer = default_serializer or BinarySerializer()

    @staticmethod
    def deserialize(serialized, serializer=None):
        serializer = serializer or BinarySerializer()
        if serialized:
            return serializer.deserialize(serialized)
        else:
            raise MacaroonInitException(
                'Must supply serialized macaroon.'
            )

    @property
    def location(self):
        return convert_to_string(self._location)

    @location.setter
    def location(self, string_or_bytes):
        self._location = convert_to_bytes(string_or_bytes)

    @property
    def identifier(self):
        return convert_to_string(self._identifier)

    @identifier.setter
    def identifier(self, string_or_bytes):
        self._identifier = convert_to_bytes(string_or_bytes)

    @property
    def signature(self):
        return convert_to_string(self._signature)

    @signature.setter
    def signature(self, string_or_bytes):
        self._signature = convert_to_bytes(string_or_bytes)

    @property
    def caveats(self):
        return self._caveats

    @caveats.setter
    def caveats(self, caveats):
        self._caveats = caveats

    def copy(self):
        return copy.deepcopy(self)

    def serialize(self, serializer=None):
        serializer = serializer or self.serializer
        return serializer.serialize(self)

    def inspect(self):
        combined = 'location {loc}\n'.format(loc=self.location)
        combined += 'identifier {id}\n'.format(id=self.identifier)
        for caveat in self.caveats:
            combined += 'cid {cid}\n'.format(cid=caveat.caveatId)
            if caveat.verificationKeyId and caveat.location:
                combined += 'vid {vid}\n'.format(vid=caveat.verificationKeyId)
                combined += 'cl {cl}\n'.format(cl=caveat.location)
        combined += 'signature {sig}'.format(sig=self.signature)
        return combined

    def first_party_caveats(self):
        return [caveat for caveat in self.caveats if caveat.first_party()]

    def third_party_caveats(self):
        return [caveat for caveat in self.caveats if caveat.third_party()]

    # Protects discharge macaroons in the event they are sent to
    # the wrong location by binding to the root macaroon
    def prepare_for_request(self, macaroon, binder_class=None):
        binder_class = binder_class or self.binder_class
        protected = macaroon.copy()
        return binder_class(self).bind(protected)

    # The existing macaroon signature is the key for hashing the
    # caveat being added. This new hash becomes the signature of
    # the macaroon with caveat added.
    def add_first_party_caveat(self, predicate):
        predicate = convert_to_bytes(predicate)
        caveat = Caveat(caveatId=convert_to_bytes(predicate))
        self.caveats.append(caveat)
        encode_key = binascii.unhexlify(self._signature)
        self._signature = sign_first_party_caveat(encode_key, predicate)
        return self

    # The third party caveat key is encrypted useing the current signature, and
    # the caveat is added to the list. The existing macaroon signature
    # is the key for hashing the string (verificationId + caveatId).
    # This new hash becomes the signature of the macaroon with caveat added.
    def add_third_party_caveat(self, location, key, key_id, nonce=None):
        derived_key = truncate_or_pad(generate_derived_key(convert_to_bytes(
            key)))
        old_key = truncate_or_pad(binascii.unhexlify(self._signature))
        box = SecretBox(key=old_key)
        encrypted = box.encrypt(derived_key, nonce=nonce)
        verificationKeyId = standard_b64encode(encrypted)
        caveat = Caveat(
            caveatId=key_id,
            location=location,
            verificationKeyId=verificationKeyId
        )
        self._caveats.append(caveat)
        encode_key = binascii.unhexlify(self._signature)
        self._signature = sign_third_party_caveat(
            encode_key,
            caveat._verificationKeyId,
            caveat._caveatId
        )
        return self



import binascii
from base64 import standard_b64decode

try:
    # py2.7.7+ and py3.3+ have native comparison support
    from hmac import compare_digest as constant_time_compare
except ImportError:
    from pymacaroons.utils import equals as constant_time_compare

from libnacl.secret import SecretBox

from pymacaroons.binders import HashSignaturesBinder
from pymacaroons.exceptions import (MacaroonInvalidSignatureException,
                                    MacaroonUnmetCaveatException)
from pymacaroons.utils import (convert_to_bytes,
                               convert_to_string,
                               truncate_or_pad,
                               generate_derived_key,
                               hmac_digest,
                               sign_first_party_caveat,
                               sign_third_party_caveat)


class Verifier(object):

    def __init__(self, default_binder_class=None, discharge_binders=None):
        self.predicates = []
        self.callbacks = []
        self.calculated_signature = None
        self.binder_class = default_binder_class or HashSignaturesBinder
        self.discharge_binders = discharge_binders or {}

    def satisfy_exact(self, predicate):
        if predicate is None:
            raise TypeError('Predicate cannot be none.')
        self.predicates.append(convert_to_string(predicate))

    def satisfy_general(self, func):
        if not hasattr(func, '__call__'):
            raise TypeError('General caveat verifiers must be callable.')
        self.callbacks.append(func)

    def verify(self, macaroon, key, discharge_macaroons=None):
        key = generate_derived_key(convert_to_bytes(key))
        return self.verify_discharge(
            macaroon,
            macaroon,
            key,
            discharge_macaroons
        )

    def verify_discharge(self, root, discharge, key,
                         discharge_macaroons=None, binder_class=None):
        self.calculated_signature = hmac_digest(
            key, discharge._identifier
        )

        self._verify_caveats(discharge, discharge_macaroons)

        if root != discharge:
            binder_class = binder_class or self.binder_class
            self.calculated_signature = binascii.unhexlify(
                binder_class(root).bind_signature(
                    binascii.hexlify(self.calculated_signature)
                )
            )

        if not self._signatures_match(
            discharge.signature,
            binascii.hexlify(self.calculated_signature)
        ):
            raise MacaroonInvalidSignatureException('Signatures do not match.')

        return True

    def _verify_caveats(self, macaroon, discharge_macaroons):
        for caveat in macaroon.caveats:
            if caveat.first_party():
                caveatMet = self._verify_first_party_caveat(
                    caveat
                )
            else:
                caveatMet = self._verify_third_party_caveat(
                    caveat,
                    macaroon,
                    discharge_macaroons
                )

            if not caveatMet:
                raise MacaroonUnmetCaveatException(
                    'Caveat not met. Unable to satisify: ' + caveat.caveatId
                )

    def _verify_first_party_caveat(self, caveat):
        caveatMet = False
        if caveat.caveatId in self.predicates:
            caveatMet = True
        else:
            for callback in self.callbacks:
                if callback(caveat.caveatId):
                    caveatMet = True
        if caveatMet:
            encode_key = self.calculated_signature
            self.calculated_signature = binascii.unhexlify(
                sign_first_party_caveat(
                    encode_key,
                    caveat._caveatId
                )
            )
        return caveatMet

    def _verify_third_party_caveat(self,
                                   caveat,
                                   root_macaroon,
                                   discharge_macaroons):
        caveatMet = False

        caveat_macaroon = \
            next((m for m in discharge_macaroons
                  if m.identifier == caveat.caveatId), None)

        if not caveat_macaroon:
            raise MacaroonUnmetCaveatException(
                'Caveat not met. No discharge macaroon found for identifier: '
                + caveat.caveatId
            )

        caveat_key = self._extract_caveat_key(caveat)

        caveat_macaroon_verifier = Verifier()
        caveat_macaroon_verifier.predicates = self.predicates
        caveat_macaroon_verifier.callbacks = self.callbacks

        caveatMet = caveat_macaroon_verifier.verify_discharge(
            root_macaroon,
            caveat_macaroon,
            caveat_key,
            discharge_macaroons=discharge_macaroons,
            binder_class=(self.discharge_binders.get(caveat.caveatId) or
                          self.discharge_binders.get(caveat.location))
        )
        if caveatMet:
            encode_key = self.calculated_signature
            self.calculated_signature = binascii.unhexlify(
                sign_third_party_caveat(
                    encode_key,
                    caveat._verificationKeyId,
                    caveat._caveatId
                )
            )
        return caveatMet

    def _extract_caveat_key(self, caveat):
        key = truncate_or_pad(self.calculated_signature)
        box = SecretBox(key=key)
        decoded_vid = standard_b64decode(
            caveat._verificationKeyId
        )
        decrypted = box.decrypt(decoded_vid)
        return decrypted

    def _signatures_match(self, s1, s2):
        sig1 = convert_to_string(s1)
        sig2 = convert_to_string(s2)
        return constant_time_compare(sig1, sig2)
